from tests import project_dir

content_names = ["content_path"]
content_paths = (
    [project_dir / "assets/data/avatar.page"],
    [project_dir / "assets/data/shannara-chronicles.page"],
)
# TODO: Update these contents in path regularly


def read_content(path):
    with open(path, encoding="utf-8") as fh:
        return fh.read()



from moviebox_api.core import Search, SubjectType
from moviebox_api.download import (
    CaptionFileDownloader,
    DownloadableMovieFilesDetail,
    DownloadableTVSeriesFilesDetail,
    MediaFileDownloader,
)
from moviebox_api.models import DownloadableFilesMetadata, SearchResultsModel
from moviebox_api.requests import Session


def test_download_movie_caption_file():
    session = Session()
    search = Search(session, "avatar", subject_type=SubjectType.MOVIES)
    search_results: SearchResultsModel = search.get_content_model_sync()
    target_movie = search_results.first_item

    downloadable_files = DownloadableMovieFilesDetail(session, target_movie)
    downloadable_files_detail: DownloadableFilesMetadata = downloadable_files.get_content_model_sync()
    target_caption_file = downloadable_files_detail.english_subtitle_file

    caption_file_downloader = CaptionFileDownloader()
    response = caption_file_downloader.run_sync(
        target_caption_file, filename=target_movie.title + "- English.srt", test=True
    )
    assert response.is_success


def test_download_movie_file():
    session = Session()
    search = Search(session, "avatar", subject_type=SubjectType.MOVIES)
    search_results: SearchResultsModel = search.get_content_model_sync()
    target_movie = search_results.first_item

    downloadable_files = DownloadableMovieFilesDetail(session, target_movie)
    downloadable_files_detail: DownloadableFilesMetadata = downloadable_files.get_content_model_sync()
    target_media_file = downloadable_files_detail.best_media_file

    media_file_downloader = MediaFileDownloader()
    response = media_file_downloader.run_sync(
        target_media_file, filename=target_movie.title + ".mp4", test=True
    )
    assert response.is_success


def test_download_tv_series_caption_file():
    session = Session()
    search = Search(session, "Merlin", subject_type=SubjectType.TV_SERIES)
    search_results: SearchResultsModel = search.get_content_model_sync()
    target_series = search_results.first_item

    downloadable_files = DownloadableTVSeriesFilesDetail(session, target_series)
    downloadable_files_detail: DownloadableFilesMetadata = downloadable_files.get_content_model_sync(
        season=1, episode=1
    )
    target_caption_file = downloadable_files_detail.english_subtitle_file

    caption_file_downloader = CaptionFileDownloader()
    response = caption_file_downloader.run_sync(target_caption_file, filename=target_series, test=True)
    assert response.is_success


def test_download_tv_series_file():
    session = Session()
    search = Search(session, "Merlin", subject_type=SubjectType.TV_SERIES)
    search_results: SearchResultsModel = search.get_content_model_sync()
    target_series = search_results.first_item

    downloadable_files = DownloadableTVSeriesFilesDetail(session, target_series)
    downloadable_files_detail: DownloadableFilesMetadata = downloadable_files.get_content_model_sync(
        season=1, episode=1
    )
    target_media_file = downloadable_files_detail.best_media_file

    media_file_downloader = MediaFileDownloader()
    response = media_file_downloader.run_sync(target_media_file, filename=target_series, test=True)
    assert response.is_success



import pytest
from pydantic import BaseModel

from moviebox_api import Search, Session, SubjectType
from moviebox_api.core import (
    HotMoviesAndTVSeries,
    PopularSearch,
    Recommend,
    SearchSuggestion,
    Trending,
)
from tests import MOVIE_KEYWORD


def test_popular_search():
    search = PopularSearch(Session())
    assert type(search.get_content_sync()) is list
    modelled_content = search.get_content_model_sync()
    assert isinstance(modelled_content[0], BaseModel)


def test_recommend():
    session = Session()

    search = Search(session, query=MOVIE_KEYWORD, subject_type=SubjectType.MOVIES)
    search_results = search.get_content_model_sync()
    target_item = search_results.first_item

    recommend = Recommend(session, item=target_item)
    recommendeded_details = recommend.get_content_model_sync()

    assert isinstance(recommendeded_details, BaseModel)
    next_recommend = recommend.next_page(recommendeded_details)
    assert isinstance(next_recommend, Recommend)

    next_recommended_details = next_recommend.get_content_model_sync()
    assert isinstance(next_recommended_details, BaseModel)
    previous_recommend = next_recommend.previous_page(next_recommended_details)

    assert isinstance(previous_recommend, Recommend)
    assert next_recommend._page > recommend._page
    assert previous_recommend._page == recommend._page


def test_search_suggestion():
    suggestion = SearchSuggestion(Session())
    suggestion_details = suggestion.get_content_model_sync(MOVIE_KEYWORD)
    assert isinstance(suggestion_details, BaseModel)


def test_hot_movies_and_series():
    hot_movies_and_series = HotMoviesAndTVSeries(Session())
    hot_item_details = hot_movies_and_series.get_content_model_sync()
    assert isinstance(hot_item_details, BaseModel)


def test_trending():
    trending = Trending(Session())

    trending_items = trending.get_content_model_sync()
    assert isinstance(trending_items, BaseModel)

    next_trends = trending.next_page(trending_items)
    assert isinstance(next_trends, Trending)
    assert next_trends._page > trending._page

    next_trending_items = next_trends.get_content_model_sync()
    assert isinstance(next_trending_items, BaseModel)

    previous_trends = next_trends.previous_page(next_trending_items)
    assert isinstance(previous_trends, Trending)
    assert previous_trends._page == trending._page

    previous_trending_items = previous_trends.get_content_model_sync()
    assert isinstance(previous_trending_items, BaseModel)



"""
Pydantic models.
"""

from dataclasses import dataclass
from datetime import date
from json import loads
from uuid import UUID

from pydantic import BaseModel, Field, HttpUrl, field_validator

from moviebox_api.constants import ITEM_DETAILS_PATH, DownloadQualitiesType, SubjectType
from moviebox_api.exceptions import ZeroMediaFileError
from moviebox_api.helpers import get_file_extension


@dataclass(frozen=True)
class MovieboxAppInfo:
    """This data is fetched when requesting for cookies,
    so I just find it important that I expose it in the package
    """

    channelType: str
    pkgName: str
    url: str
    versionCode: str
    versionName: str


class ContentImageModel(BaseModel):
    """Model for content image"""

    url: HttpUrl
    width: int
    height: int
    size: int
    format: str
    thumbnail: str
    blurHash: str
    gif: str | None = None
    avgHueLight: str
    avgHueDark: str
    id: str


class ContentSubjectModel(BaseModel):
    subjectId: str
    subjectType: SubjectType
    title: str
    description: str
    releaseDate: date
    duration: int
    genre: list[str]
    cover: ContentImageModel
    countryName: str
    imdbRatingValue: float
    # subtitles : str
    # ops : {rid: uuid, trace_id: str}
    # hasResource :bool
    trailer: str | None = None
    detailPath: str
    stafflist: list | None = None
    appointmentCnt: int
    appointmentDate: str
    corner: str
    # imdbRatingCount: int

    @field_validator("genre", mode="before")
    def validate_genre(value: str) -> list[str]:
        return value.split(",")


class ContentModel(BaseModel):
    """Model for a particular movie or tv series"""

    id: str
    title: str
    image: ContentImageModel
    url: HttpUrl
    subjectId: str
    subjectType: SubjectType
    subject: ContentSubjectModel | None = None

    @property
    def is_movie(self) -> bool:
        """Check whether content is a movie._"""
        return self.subjectType == SubjectType.MOVIES.value

    @property
    def is_tv_series(self) -> bool:
        """Check whether content is a tv series."""
        return self.subjectType == SubjectType.TV_SERIES.value

    @property
    def is_music(self) -> bool:
        """Check whether content is a music"""
        return self.subjectType == SubjectType.MUSIC.value


class PlatformsModel(BaseModel):
    name: str
    uploadBy: str


class ContentCategoryBannerModel(BaseModel):
    items: list[ContentModel]  # list of series/movies


class ContentCategorySubjectsModel(ContentSubjectModel):
    # also better called operatingListSubjects
    subtitles: list[str]
    ops: str
    hasResource: bool

    @field_validator("subtitles", mode="before")
    def validate_subtitles(value: str) -> list[str]:
        return value.split(",")


class ContentCategoryModel(BaseModel):
    # named: OperatingList in server response
    type: str
    position: int
    title: str
    subjects: list[ContentCategorySubjectsModel]
    banner: ContentCategoryBannerModel | None = None
    opId: str
    url: str
    livelist: list | None = None


class HomepageContentModel(BaseModel):
    """Main model for home contents

    - Movies/series available under path `operatingList[0].banner.items`
    """

    topPickList: list
    homeList: list
    url: str
    referer: str
    allPlatform: list
    banner: str | None = None
    live: str | None = None
    platformList: list[PlatformsModel]
    shareParam: str | None = None
    operatingList: list[ContentCategoryModel]

    @property
    def contents(self) -> list[ContentModel]:
        """Both movies and tv series"""
        cached_contents = []
        for operating in self.operatingList:
            if operating.banner is not None:
                cached_contents.extend(operating.banner.items)
        return cached_contents


class OPS(BaseModel):
    """A value in specific result info"""

    rid: UUID
    trace_id: str


class SearchResultsItem(ContentSubjectModel):
    """Specific result info"""

    subtitles: list[str]
    ops: OPS
    hasResource: bool
    imdbRatingCount: int | None = None  # None for TrendingResults

    @field_validator("ops", mode="before")
    def validate_ops(value: str) -> dict:
        return loads(value)

    @field_validator("subtitles", mode="before")
    def validate_subtitles(value: str) -> list[str]:
        return value.split(",")

    @property
    def page_url(self) -> str:
        """Url to the specific item details page"""
        return f"{ITEM_DETAILS_PATH}/{self.detailPath}?id={self.subjectId}"


class SearchResultsPagerModel(BaseModel):
    """Search pagination info"""

    hasMore: bool
    nextPage: int
    page: int
    perPage: int
    totalCount: int


class SearchResultsModel(BaseModel):
    """Whole search results"""

    pager: SearchResultsPagerModel
    items: list[SearchResultsItem]

    @property
    def first_item(self) -> SearchResultsItem:
        return self.items[0]


class TrendingResultsModel(BaseModel):
    """Whole trending results"""

    pager: SearchResultsPagerModel
    subjectList: list[SearchResultsItem]

    @property
    def items(self) -> list[SearchResultsItem]:
        return self.subjectList

    @property
    def first_item(self) -> SearchResultsItem:
        return self.items[0]


class HotMoviesAndTVSeriesModel(BaseModel):
    movies: list[SearchResultsItem] = Field(alias="movie")
    tv_series: list[SearchResultsItem] = Field(alias="tv")


class SuggestedItemsModel(BaseModel):
    """Items suggested"""

    items: list
    keyword: str
    ops: str


class BaseFileMetadata(BaseModel):
    @property
    def ext(self) -> str:
        """Media file extension such as `mp4` or `srt`"""
        return get_file_extension(self.url)


class MediaFileMetadata(BaseFileMetadata):
    id: str
    url: HttpUrl
    resolution: int
    size: int


class CaptionFileMetadata(BaseFileMetadata):
    id: str
    lan: str
    lanName: str
    url: HttpUrl
    size: int
    delay: int


class DownloadableFilesMetadata(BaseModel):
    downloads: list[MediaFileMetadata]
    captions: list[CaptionFileMetadata]
    limited: bool
    limitedCode: str
    hasResource: bool

    def _check_downloads(self) -> bool:
        """Checks whether there are downloadable media file.

        Raises:
            ZeroMediaFileError: Incase the downloads list is empty

        Returns:
            bool: Downloadable media file(s) exist.
        """
        if bool(self.downloads):
            return True

        raise ZeroMediaFileError("There are no downloadable  mediafiles for the targeted item")

    @property
    def best_media_file(self) -> MediaFileMetadata:
        """Highest quality media file"""
        self._check_downloads()
        if bool(self.downloads):
            found = self.downloads[0]
            for media_file in self.downloads[1:]:
                if media_file.resolution > found.resolution:
                    found = media_file
            return found

    @property
    def worst_media_file(self) -> MediaFileMetadata:
        """Lowest quality media file"""
        self._check_downloads()
        if bool(self.downloads):
            found = self.downloads[0]
            for media_file in self.downloads[1:]:
                if media_file.resolution < found.resolution:
                    found = media_file
            return found

    @property
    def english_subtitle_file(self) -> CaptionFileMetadata | None:
        """English subtitle file."""
        for subtitle_file in self.captions:
            if subtitle_file.lan == "en":
                return subtitle_file

    def get_quality_downloads_map(
        self,
    ) -> dict[DownloadQualitiesType, MediaFileMetadata]:
        """Maps media file qualities to their equivalent media files object

        Returns:
            dict[DownloadQualitiesType, MediaFileMetadata]
        """
        resolution_downloads_map = {}
        for item in self.downloads:
            resolution_downloads_map[f"{item.resolution}P"] = item
        return resolution_downloads_map

    def get_media_file_by_resolution(self, resolution: int) -> MediaFileMetadata:
        """Get specific MediaFileMetadata based on resolution.

        Args:
            resolution (int): Media file resolution e.g 480, 720, 1080 etc

        Returns:
            MediaFileMetadata: Media file matching that resolution.

        Raises:
            ValueError: Incase no media_file matched the resolution.
        """
        available_media_file_resolutions = []
        for media_file in self.downloads:
            available_media_file_resolutions.append(media_file.resolution)
            if media_file.resolution == resolution:
                return media_file
        raise ValueError(
            "No media_file matched that resolution. Available resolutions "
            f"include {available_media_file_resolutions}"
        )

    def get_language_subtitle_map(
        self,
    ) -> dict[str, CaptionFileMetadata]:
        """Returns something like { English : CaptionFileMetadata }"""
        language_subtitle_map = {}
        for caption in self.captions:
            language_subtitle_map[caption.lanName] = caption
        return language_subtitle_map

    def get_language_short_subtitle_map(
        self,
    ) -> dict[str, CaptionFileMetadata]:
        """Returns something like { en : CaptionFileMetadata }"""
        language_subtitle_map = {}
        for caption in self.captions:
            language_subtitle_map[caption.lan] = caption
        return language_subtitle_map

    def get_subtitle_by_language(self, language: str) -> CaptionFileMetadata | None:
        """Both `English` and `en` will return same thing"""
        if len(language) == 2:
            return self.get_language_short_subtitle_map().get(language.lower())
        return self.get_language_subtitle_map().get(language.capitalize())


class StreamFileMetadata(BaseModel):
    format: str
    id: str
    url: HttpUrl
    resolutions: int
    size: int
    duration: int
    codecName: str


class StreamFilesMetadata(BaseModel):
    streams: list[StreamFileMetadata]
    freeNum: int
    limited: bool
    dash: list
    hls: list
    hasResource: bool

    @property
    def best_stream_file(self) -> StreamFileMetadata | None:
        """Highest quality stream file"""
        if bool(self.streams):
            found = self.streams[0]
            for stream_file in self.streams[1:]:
                if stream_file.resolutions > found.resolutions:
                    found = stream_file
            return found

    @property
    def worst_stream_file(self) -> StreamFileMetadata | None:
        """Lowest quality stream file"""
        if bool(self.streams):
            found = self.streams[0]
            for stream_file in self.streams[1:]:
                if stream_file.resolutions < found.resolutions:
                    found = stream_file
            return found


class PopularSearchModel(BaseModel):
    """Item many people are searching"""

    title: str



import pytest
from pydantic import BaseModel

from moviebox_api.extractor._core import (
    JsonDetailsExtractorModel,
    TagDetailsExtractorModel,
)
from tests.extractors import (
    content_names,
    content_paths,
    read_content,
)


@pytest.mark.parametrize(content_names, content_paths)
def test_json_details_extractor_model(content_path):
    content = read_content(content_path)
    extractor = JsonDetailsExtractorModel(content)
    assert isinstance(extractor.details, BaseModel)
    assert isinstance(extractor.data, BaseModel)
    assert isinstance(extractor.subject, BaseModel)
    assert isinstance(extractor.metadata, BaseModel)

    assert isinstance(extractor.resource, BaseModel)
    assert isinstance(extractor.reviews[0], BaseModel)
    assert isinstance(extractor.seasons[0], BaseModel)

    assert isinstance(extractor.stars[0], BaseModel)
    assert isinstance(extractor.page_details, BaseModel)


@pytest.mark.parametrize(content_names, content_paths)
def test_tag_details_extractor_model(content_path):
    content = read_content(content_path)
    extractor = TagDetailsExtractorModel(content)
    assert isinstance(extractor.extract_headers(), BaseModel)
    assert isinstance(extractor.extract_basics(), BaseModel)
    assert isinstance(extractor.extract_casts()[0], BaseModel)

    assert isinstance(extractor.extract_reviews()[0], BaseModel)
    assert isinstance(extractor.extract_others(), BaseModel)



"""Gets the work done - downloads media with flexible flow control"""

import logging
import tempfile
from pathlib import Path
from typing import Literal

import httpx
from throttlebuster import DownloadedFile
from throttlebuster.constants import DOWNLOAD_PART_EXTENSION

from moviebox_api.cli.helpers import (
    get_caption_file_or_raise,
    media_player_name_func_map,
    perform_search_and_get_item,
)
from moviebox_api.constants import (
    CURRENT_WORKING_DIR,
    DEFAULT_CAPTION_LANGUAGE,
    DEFAULT_CHUNK_SIZE,
    DEFAULT_TASKS,
    DOWNLOAD_QUALITIES,
    DownloadQualitiesType,
    SubjectType,
)
from moviebox_api.core import Session, TVSeriesDetails
from moviebox_api.download import (
    CaptionFileDownloader,
    DownloadableMovieFilesDetail,
    DownloadableTVSeriesFilesDetail,
    MediaFileDownloader,
    resolve_media_file_to_be_downloaded,
)
from moviebox_api.exceptions import ZeroCaptionFileError
from moviebox_api.helpers import assert_instance, assert_membership, get_event_loop
from moviebox_api.models import SearchResultsItem

__all__ = ["Downloader"]


class Downloader:
    """Controls the movie/series download process"""

    def __init__(self, session: Session = None):
        """Constructor for `Downloader`

        Args:
            session (Session, optional): MovieboxAPI httpx request session . Defaults to Session().
        """
        self._session = session if session else Session()
        assert_instance(self._session, Session, "session")

    async def download_movie(
        self,
        title: str,
        year: int | None = None,
        yes: bool = False,
        dir: Path | str = CURRENT_WORKING_DIR,
        caption_dir: Path | str = CURRENT_WORKING_DIR,
        quality: DownloadQualitiesType = "BEST",
        movie_filename_tmpl: str = MediaFileDownloader.movie_filename_template,
        caption_filename_tmpl: str = CaptionFileDownloader.movie_filename_template,
        language: tuple[str] = (DEFAULT_CAPTION_LANGUAGE,),
        download_caption: bool = False,
        caption_only: bool = False,
        stream_via: Literal["mpv", "vlc"] | None = None,
        search_function: callable = perform_search_and_get_item,
        chunk_size: int = DEFAULT_CHUNK_SIZE,
        tasks: int = DEFAULT_TASKS,
        part_dir: Path | str = CURRENT_WORKING_DIR,
        part_extension: str = DOWNLOAD_PART_EXTENSION,
        merge_buffer_size: int | None = None,
        ignore_missing_caption: bool = False,
        **run_kwargs,
    ) -> tuple[
        DownloadedFile | httpx.Response | None,
        list[DownloadedFile | httpx.Response] | None,
    ]:
        """Search movie by name and proceed to download it or stream it.

        Args:
            title (str): Complete or partial movie name
            year (int|None, optional): `releaseDate.year` filter for the movie. Defaults to None.
            yes (bool, optional): Proceed with the first item in the results instead of prompting confirmation. Defaults to False
            dir (Path|str, optional): Directory for saving the movie file to. Defaults to CURRENT_WORKING_DIR.
            caption_dir (Path|str, optional): Directory for saving the caption file to. Defaults to CURRENT_WORKING_DIR.
            quality (DownloadQualitiesType, optional): Such as `720p` or simply `BEST` etc. Defaults to 'BEST'.
            movie_filename_tmpl (str, optional): Template for generating movie filename. Defaults to MediaFileDownloader.movie_filename_template.
            caption_filename_tmpl (str, optional): Template for generating caption filename. Defaults to CaptionFileDownloader.movie_filename_template.
            language (tuple, optional): Languages to download captions in. Defaults to (DEFAULT_CAPTION_LANGUAGE,).
            download_caption (bool, optional): Whether to download caption or not. Defaults to False.
            caption_only (bool, optional): Whether to ignore movie file or not. Defaults to False.
            stream_via (Literal["mpv", "vlc"] | None = None, optional): Stream directly in chosen media_player instead of downloading. Defaults to None.
            search_function (callable, optional): Accepts `session`, `title`, `year`, `subject_type` & `yes` and returns `SearchResultsItem`.
            chunk_size (int, optional): Streaming download chunk size in kilobytes. Defaults to DEFAULT_CHUNK_SIZE.
            tasks (int, optional): Number of tasks to carry out the download. Defaults to DEFAULT_TASKS.
            part_dir (Path | str, optional): Directory for temporarily saving the downloaded file-parts to. Defaults to CURRENT_WORKING_DIR.
            part_extension (str, optional): Filename extension for download parts. Defaults to DOWNLOAD_PART_EXTENSION.
            merge_buffer_size (int|None, optional). Buffer size for merging the separated files in kilobytes. Defaults to chunk_size.

        run_kwargs: Other keyword arguments for `MediaFileDownloader.run`

        Returns:
            tuple[DownloadedFile | httpx.Response  | None, list[DownloadedFile | httpx.Response ] | None]: Path to downloaded movie and downloaded caption files.
        """  # noqa: E501

        assert_membership(quality, DOWNLOAD_QUALITIES)

        assert callable(search_function), (
            f"Value for search_function must be callable not {type(search_function)}"
        )

        MediaFileDownloader.movie_filename_template = movie_filename_tmpl
        CaptionFileDownloader.movie_filename_template = caption_filename_tmpl

        target_movie = await search_function(
            self._session,
            title=title,
            year=year,
            subject_type=SubjectType.MOVIES,
            yes=yes,
        )

        assert isinstance(target_movie, SearchResultsItem), (
            f"Search function {search_function.__name__} must return an instance of "
            f"{SearchResultsItem} not {type(target_movie)}"
        )

        downloadable_details_inst = DownloadableMovieFilesDetail(self._session, target_movie)

        downloadable_details = await downloadable_details_inst.get_content_model()

        target_media_file = resolve_media_file_to_be_downloaded(quality, downloadable_details)

        subtitle_details_items: list[DownloadedFile] = []

        subtitles_dir = tempfile.mkdtemp() if stream_via else caption_dir

        if download_caption or caption_only:
            for lang in language:
                try:
                    target_caption_file = get_caption_file_or_raise(downloadable_details, lang)

                except (ZeroCaptionFileError, ValueError):
                    if ignore_missing_caption:
                        continue
                    raise

                caption_downloader = CaptionFileDownloader(
                    dir=subtitles_dir,
                    chunk_size=chunk_size,
                    tasks=tasks,
                    part_dir=part_dir,
                    part_extension=part_extension,
                    merge_buffer_size=merge_buffer_size,
                )

                subtitle_details = await caption_downloader.run(
                    caption_file=target_caption_file,
                    filename=target_movie,
                    **run_kwargs,
                )

                subtitle_details_items.append(subtitle_details)

            if caption_only and not stream_via:
                # terminate
                return (None, subtitle_details_items)

        if stream_via:
            return media_player_name_func_map[stream_via](
                str(target_media_file.url), subtitle_details_items, subtitles_dir
            )

        movie_downloader = MediaFileDownloader(
            dir=dir,
            chunk_size=chunk_size,
            tasks=tasks,
            part_dir=part_dir,
            part_extension=part_extension,
            merge_buffer_size=merge_buffer_size,
        )

        movie_details = await movie_downloader.run(
            media_file=target_media_file, filename=target_movie, **run_kwargs
        )
        return (movie_details, subtitle_details_items)

    async def download_tv_series(
        self,
        title: str,
        season: int,
        episode: int,
        year: int | None = None,
        yes: bool = False,
        dir: Path | str = CURRENT_WORKING_DIR,
        caption_dir: Path | str = CURRENT_WORKING_DIR,
        quality: DownloadQualitiesType = "BEST",
        episode_filename_tmpl: str = MediaFileDownloader.series_filename_template,
        caption_filename_tmpl: str = CaptionFileDownloader.series_filename_template,
        language: tuple = (DEFAULT_CAPTION_LANGUAGE,),
        download_caption: bool = False,
        caption_only: bool = False,
        stream_via: Literal["mpv", "vlc"] | None = None,
        limit: int = 1,
        search_function: callable = perform_search_and_get_item,
        chunk_size: int = DEFAULT_CHUNK_SIZE,
        tasks: int = DEFAULT_TASKS,
        part_dir: Path | str = CURRENT_WORKING_DIR,
        part_extension: str = DOWNLOAD_PART_EXTENSION,
        merge_buffer_size: int | None = None,
        ignore_missing_caption: bool = False,
        auto_mode: bool = False,
        format: Literal["group", "struct"] | None = None,
        **run_kwargs,
    ) -> dict[
        int,
        dict[
            int,
            dict[str, DownloadedFile | httpx.Response | list[DownloadedFile | httpx.Response]],
        ],
    ]:
        """Search tv-series by name and proceed to download or stream its episodes.

        Args:
            title (str): Complete or partial tv-series name.
            season (int): Target season number of the tv-series.
            episode (int): Target episode number of the tv-series.
            year (int|None, optional): `releaseDate.year` filter for the tv-series. Defaults to None.
            yes (bool, optional): Proceed with the first item in the results instead of prompting confirmation. Defaults to False.
            dir (Path|str, optional): Directory for saving the movie file to. Defaults to CURRENT_WORKING_DIR.
            caption_dir (Path|str, optional): Directory for saving the caption files to. Defaults to CURRENT_WORKING_DIR.
            quality (DownloadQualitiesType, optional): Episode quality such as `720p` or simply `BEST` etc. Defaults to 'BEST'.
            episode_filename_tmpl (str, optional): Template for generating episode filename. Defaults to MediaFileDownloader.series_filename_template.
            caption_filename_tmpl (str, optional): Template for generating caption filename. Defaults to CaptionFileDownloader.series_filename_template.
            language (tuple, optional): Languages to download captions in. Defaults to (DEFAULT_CAPTION_LANGUAGE,).
            download_caption (bool, optional): Whether to download caption or not. Defaults to False.
            caption_only (bool, optional): Whether to ignore episode files or not. Defaults to False.
            stream_via (Literal["mpv", "vlc"], optional): Stream directly in chosen media played instead of downloading. Defaults to None.
            limit (int, optional): Number of episodes to download including the offset episode. Defaults to 1.
            search_function (callable, optional): Accepts `session`, `title`, `year`, `subject_type` & `yes` and returns item.
            chunk_size (int, optional): Streaming download chunk size in kilobytes. Defaults to DEFAULT_CHUNK_SIZE.
            tasks (int, optional): Number of tasks to carry out the download. Defaults to DEFAULT_TASKS.
            part_dir (Path | str, optional): Directory for temporarily saving the downloaded file-parts to. Defaults to CURRENT_WORKING_DIR.
            part_extension (str, optional): Filename extension for download parts. Defaults to DOWNLOAD_PART_EXTENSION.
            merge_buffer_size (int|None, optional). Buffer size for merging the separated files in kilobytes. Defaults to chunk_size.
            auto_mode (bool, optional). Iterate over seasons as well. When limit is 1 (default), download entire tv series. Defaults to False.
            format(Literal["filename", "group", "struct"] | None, optional): Ways of formating filename and saving the episodes. Defaults to None
                group -> Organize episodes into separate folders based on seasons e.g Merlin/S1/Merlin S1E2.mp4
                struct -> Save episodes in a hierarchical directory structure e.g Merlin (2009)/S1/E1.mp4

        run_kwargs: Other keyword arguments for `MediaFileDownloader.run`

        Returns:
             dict[int, dict[str, DownloadedFile | httpx.Response  | list[DownloadedFile | httpx.Response ]]]: Episode number and downloaded episode file details and caption files.
        """  # noqa: E501

        assert_membership(quality, DOWNLOAD_QUALITIES)

        assert callable(search_function), (
            f"Value for search_function must be callable not {type(search_function)}"
        )

        MediaFileDownloader.series_filename_template = episode_filename_tmpl
        CaptionFileDownloader.series_filename_template = caption_filename_tmpl

        group = False

        match format:
            case "group":
                group = True

            case "struct":
                MediaFileDownloader.series_filename_template = "E{episode}.{ext}"
                CaptionFileDownloader.series_filename_template = "E{episode}.{lan}.{ext}"
                group = True

        target_tv_series = await search_function(
            self._session,
            title=title,
            year=year,
            subject_type=SubjectType.TV_SERIES,
            yes=yes,
        )
        assert isinstance(target_tv_series, SearchResultsItem), (
            f"Search function {search_function.__name__} must return an instance of "
            f"{SearchResultsItem} not {type(target_tv_series)}"
        )

        downloadable_files = DownloadableTVSeriesFilesDetail(self._session, target_tv_series)

        subtitles_dir = tempfile.mkdtemp() if stream_via else caption_dir

        caption_downloader = CaptionFileDownloader(
            dir=(subtitles_dir if stream_via else dir if group else subtitles_dir),
            chunk_size=chunk_size,
            tasks=tasks,
            part_dir=part_dir,
            part_extension=part_extension,
            merge_buffer_size=merge_buffer_size,
            group_series=group,
        )

        media_file_downloader = MediaFileDownloader(
            dir=dir,
            chunk_size=chunk_size,
            tasks=tasks,
            part_dir=part_dir,
            part_extension=part_extension,
            merge_buffer_size=merge_buffer_size,
            group_series=group,
        )

        core_tv_series_details = TVSeriesDetails(target_tv_series, self._session)

        tv_series_details_model = await core_tv_series_details.get_json_details_extractor_model()
        series_resource = tv_series_details_model.resource

        async def download_episodes_per_season(
            season_number: int,
            first_episode_number: int,
            episode_limit: int,
        ):
            response = {}
            for episode_count in range(episode_limit):
                current_episode = first_episode_number + episode_count

                downloadable_files_detail = await downloadable_files.get_content_model(
                    season=season_number, episode=current_episode
                )

                current_episode_details = {}
                caption_details_items: list[DownloadedFile] = []

                if caption_only or download_caption:
                    for lang in language:
                        try:
                            target_caption_file = get_caption_file_or_raise(downloadable_files_detail, lang)

                        except (ZeroCaptionFileError, ValueError):
                            if ignore_missing_caption:
                                continue
                            raise

                        caption_details = await caption_downloader.run(
                            caption_file=target_caption_file,
                            filename=target_tv_series,
                            season=season_number,
                            episode=current_episode,
                            **run_kwargs,
                        )

                        caption_details_items.append(caption_details)

                    if caption_only and not stream_via:
                        # Avoid downloading tv-series
                        continue

                # Download or stream series

                current_episode_details["captions"] = caption_details_items

                target_media_file = resolve_media_file_to_be_downloaded(quality, downloadable_files_detail)

                if stream_via:
                    media_player_name_func_map[stream_via](
                        str(target_media_file.url), caption_details_items, subtitles_dir
                    )

                    continue

                tv_series_details = await media_file_downloader.run(
                    media_file=target_media_file,
                    filename=target_tv_series,
                    season=season_number,
                    episode=current_episode,
                    **run_kwargs,
                )

                current_episode_details["movie"] = tv_series_details
                response[current_episode] = current_episode_details

            return response

        if auto_mode:
            if series_resource.total_seasons < season:
                raise RuntimeError(
                    f"The target season {season} exceeds the available "
                    f"tv series seasons {series_resource.total_seasons}."
                )

            total_episodes = 0
            downloaded_episodes_count = 0
            response_jar = {}

            target_seasons = series_resource.seasons[season - 1 :]

            for index, series_season in enumerate(target_seasons):
                new_episodes_count = series_season.maxEp

                if index == 0:
                    # episode offset
                    if series_season.maxEp < episode:
                        raise RuntimeError(
                            f"The target episode offset {episode} for season {series_season.se}"
                            f" is greater than the available episodes {series_season.maxEp}"
                        )

                    else:
                        new_episodes_count -= episode - 1

                total_episodes += new_episodes_count

            if limit != 1:
                if limit > total_episodes:
                    logging.warning(
                        f"You have set total episodes limit to {limit} but only {total_episodes} "
                        f"episodes are available starting from the offset ({season=}, {episode=}"
                        "). The former will be ignored."
                    )
                    limit = total_episodes

            else:
                limit = total_episodes

            logging.info(
                f"Process overview - Seasons: {len(target_seasons)}, total episodes: {total_episodes}, "
                f"episodes download limit: {limit} "
            )

            for index, target_season in enumerate(target_seasons):
                if index == 0:
                    first_episode_number = episode  # declared by user
                    episodes_limit = target_season.maxEp - (episode - 1)  # 1 = index 0

                    if episodes_limit > limit:
                        episodes_limit = limit

                else:
                    first_episode_number = 1

                    remaining_episodes_amount = limit - downloaded_episodes_count

                    if target_season.maxEp > remaining_episodes_amount:
                        episodes_limit = remaining_episodes_amount

                    else:
                        episodes_limit = target_season.maxEp

                downloaded_episodes_details = await download_episodes_per_season(
                    season_number=target_season.se,
                    first_episode_number=first_episode_number,
                    episode_limit=episodes_limit,
                )
                response_jar[target_season.se] = downloaded_episodes_details

                downloaded_episodes_count += episodes_limit

            return response_jar

        else:
            target_season = series_resource.get_season_by_number(season)

            assert episode <= target_season.maxEp, (
                f"The chosen episode offset {episode} exceeds the available episodes {target_season.maxEp}"
            )

            available_episodes = target_season.maxEp - (episode - 1)  # offset

            if limit > available_episodes:
                logging.warning(
                    f"You have set episodes limit to {limit} but only {available_episodes} "
                    f"episodes are available for season {season}, starting from the offset {episode}. "
                    "The former will be ignored."
                )
                limit = available_episodes

            logging.info(
                f"Season {target_season.se} details - Total episodes: {target_season.maxEp}, "
                f"episodes download limit: {limit}"
            )

            downloaded_tv_series_details = await download_episodes_per_season(
                season_number=season, first_episode_number=episode, episode_limit=limit
            )

            return {season: downloaded_tv_series_details}

    def download_movie_sync(
        self,
        *args,
        **kwargs,
    ) -> tuple[
        DownloadedFile | httpx.Response | None,
        list[DownloadedFile | httpx.Response] | None,
    ]:
        """Synchronously search movie by name and proceed to download or stream it."""
        return get_event_loop().run_until_complete(self.download_movie(*args, **kwargs))

    def download_tv_series_sync(
        self,
        *args,
        **kwargs,
    ) -> dict[
        int,
        dict[str, DownloadedFile | httpx.Response | list[DownloadedFile | httpx.Response]],
    ]:
        """Synchronously search tv-series by name and proceed to download or stream its episodes."""
        return get_event_loop().run_until_complete(self.download_tv_series(*args, **kwargs))



"""Contains common functions for the submodule"""

from bs4 import BeautifulSoup


def souper(html: str) -> BeautifulSoup:
    """Convert html formatted txt to bts object

    Args:
        html (str): Html formatted text

    Returns:
        BeautifulSoup: Souped html
    """
    return BeautifulSoup(html, "html.parser")



"""Extracts data from specific movie/tv-series page"""

from moviebox_api.extractor._core import (
    JsonDetailsExtractor,
    JsonDetailsExtractorModel,
    TagDetailsExtractor,
    TagDetailsExtractorModel,
)
from moviebox_api.extractor.exceptions import DetailsExtractionError

__all__ = [
    "TagDetailsExtractor",
    "JsonDetailsExtractor",
    "TagDetailsExtractorModel",
    "JsonDetailsExtractorModel",
    "DetailsExtractionError",
]



"""Models for the extracted item details"""

from moviebox_api.extractor.models.json import ItemJsonDetailsModel
from moviebox_api.extractor.models.tag import ItemTagDetailsModel

__all__ = ["ItemJsonDetailsModel", "ItemTagDetailsModel"]



from moviebox_api.cli import Downloader


async def main():
    downloader = Downloader()
    episodes_content_map = await downloader.download_tv_series(
        "Merlin",
        season=1,
        episode=1,
        limit=1,
        # limit=13 # This will download entire 13 episodes of season 1
    )

    print(episodes_content_map)


if __name__ == "__main__":
    import asyncio

    asyncio.run(main())



import pytest
from pydantic import BaseModel

from moviebox_api.constants import SubjectType
from moviebox_api.core import MovieDetails, Search, TVSeriesDetails
from moviebox_api.extractor import (
    JsonDetailsExtractor,
    JsonDetailsExtractorModel,
    TagDetailsExtractor,
    TagDetailsExtractorModel,
)
from moviebox_api.requests import Session
from tests import MOVIE_KEYWORD, TV_SERIES_KEYWORD


@pytest.mark.asyncio
@pytest.mark.parametrize(
    argnames=["url"],
    argvalues=(
        ["https://moviebox.pk/detail/titanic-m7a9yt0abq6?id=5390197429792821032"],
        [
            "/detail/titanic-m7a9yt0abq6?id=5390197429792821032",
        ],
    ),
)
async def test_movie_using_page_url(url):
    session = Session()
    details = MovieDetails(
        url,
        session=session,
    )
    assert type(await details.get_html_content()) is str
    assert type(await details.get_content()) is dict
    assert isinstance(await details.get_content_model(), BaseModel)

    assert isinstance(await details.get_json_details_extractor(), JsonDetailsExtractor)
    assert isinstance(await details.get_tag_details_extractor(), TagDetailsExtractor)

    assert isinstance(await details.get_json_details_extractor_model(), JsonDetailsExtractorModel)
    assert isinstance(await details.get_tag_details_extractor_model(), TagDetailsExtractorModel)


@pytest.mark.asyncio
@pytest.mark.parametrize(
    argnames=["url"],
    argvalues=(
        [
            "https://moviebox.pk/detail/merlin-sMxCiIO6fZ9?id=8382755684005333552&scene&page_from=search_detail&type=%2Fmovie%2Fdetail"
        ],
        [
            "https://moviebox.pk/detail/merlin-sMxCiIO6fZ9?id=8382755684005333552",
        ],
    ),
)
async def test_tv_series_using_page_url(url):
    session = Session()
    details = TVSeriesDetails(
        url,
        session=session,
    )
    assert type(await details.get_html_content()) is str
    assert type(await details.get_content()) is dict
    assert isinstance(await details.get_content_model(), BaseModel)

    assert isinstance(await details.get_json_details_extractor(), JsonDetailsExtractor)
    assert isinstance(await details.get_tag_details_extractor(), TagDetailsExtractor)

    assert isinstance(await details.get_json_details_extractor_model(), JsonDetailsExtractorModel)
    assert isinstance(await details.get_tag_details_extractor_model(), TagDetailsExtractorModel)


@pytest.mark.asyncio
async def test_movie_using_search_results_item():
    session = Session()
    search = Search(session, query=MOVIE_KEYWORD, subject_type=SubjectType.MOVIES)
    search_results = await search.get_content_model()
    details = MovieDetails(
        search_results.first_item,
        session=session,
    )
    assert type(await details.get_html_content()) is str
    assert type(await details.get_content()) is dict
    assert isinstance(await details.get_content_model(), BaseModel)

    assert isinstance(await details.get_json_details_extractor(), JsonDetailsExtractor)
    assert isinstance(await details.get_tag_details_extractor(), TagDetailsExtractor)

    assert isinstance(await details.get_json_details_extractor_model(), JsonDetailsExtractorModel)
    assert isinstance(await details.get_tag_details_extractor_model(), TagDetailsExtractorModel)


@pytest.mark.asyncio
async def test_tv_series_using_search_results_item():
    session = Session()
    search = Search(
        session,
        query=TV_SERIES_KEYWORD,
        subject_type=SubjectType.TV_SERIES,
    )
    search_results = await search.get_content_model()
    details = TVSeriesDetails(
        search_results.first_item,
        session=session,
    )
    assert type(await details.get_html_content()) is str
    assert type(await details.get_content()) is dict
    assert isinstance(await details.get_content_model(), BaseModel)

    assert isinstance(await details.get_json_details_extractor(), JsonDetailsExtractor)
    assert isinstance(await details.get_tag_details_extractor(), TagDetailsExtractor)

    assert isinstance(await details.get_json_details_extractor_model(), JsonDetailsExtractorModel)
    assert isinstance(await details.get_tag_details_extractor_model(), TagDetailsExtractorModel)



"""Contains non-essential cli-commands"""

import click
import rich
from rich.table import Table

from moviebox_api.cli.helpers import (
    command_context_settings,
    perform_search_and_get_item,
    prepare_start,
)
from moviebox_api.constants import MIRROR_HOSTS, SubjectType
from moviebox_api.core import Homepage, MovieDetails, PopularSearch, TVSeriesDetails
from moviebox_api.extractor import JsonDetailsExtractor
from moviebox_api.helpers import get_event_loop
from moviebox_api.requests import Session


@click.command(context_settings=command_context_settings)
@click.option("-J", "--json", is_flag=True, help="Output details in json format")
@click.option(
    "-V",
    "--verbose",
    count=True,
    help="Show more detailed interactive texts",
    default=0,
)
@click.option(
    "-Q",
    "--quiet",
    is_flag=True,
    help="Disable showing interactive texts on the progress (logs)",
)
def mirror_hosts_command(json: bool, **start_kwargs):
    """Discover Moviebox mirror hosts [env: MOVIEBOX_API_HOST]"""
    prepare_start(**start_kwargs)

    if json:
        rich.print_json(data=dict(details=MIRROR_HOSTS), indent=4)
    else:
        table = Table(
            title="Moviebox mirror hosts",
            show_lines=True,
        )
        table.add_column("No.", style="white", justify="center")
        table.add_column("Mirror Host", style="cyan", justify="left")

        for no, mirror_host in enumerate(MIRROR_HOSTS, 1):
            table.add_row(str(no), mirror_host)

        rich.print(table)


@click.command(context_settings=command_context_settings)
@click.option(
    "-J",
    "--json",
    is_flag=True,
    help="Output details in json format : False",
)
@click.option(
    "-T",
    "--title",
    help="Title filter for the contents to list : None",
)
@click.option(
    "-B",
    "--banner",
    is_flag=True,
    help="Show banner content only : False",
)
@click.option(
    "-V",
    "--verbose",
    count=True,
    help="Show more detailed interactive texts",
    default=0,
)
@click.option(
    "-Q",
    "--quiet",
    is_flag=True,
    help="Disable showing interactive texts on the progress (logs)",
)
def homepage_content_command(json: bool, title: str, banner: bool, **start_kwargs):
    """Show contents displayed at landing page"""
    # TODO: Add automated test for this command
    prepare_start(**start_kwargs)

    session = Session()
    homepage = Homepage(session)
    homepage_contents = get_event_loop().run_until_complete(homepage.get_content_model())

    banners: dict[str, list[list[str]]] = {}
    items: dict[str, list[list[str]]] = {}

    for operating in homepage_contents.operatingList:
        if operating.type == "BANNER":
            banners[operating.title] = [
                [
                    item.subjectType.name,
                    item.title,
                    ", ".join(item.subject.genre),
                    str(item.subject.releaseDate),
                ]
                for item in operating.banner.items
                if item.subject is not None
            ]

        elif operating.type == "SUBJECTS_MOVIE":
            items[operating.title] = (
                [
                    subject.subjectType.name,
                    subject.title,
                    ", ".join(subject.genre),
                    str(subject.imdbRatingValue),
                    subject.countryName,
                    str(subject.releaseDate),
                ]
                for subject in operating.subjects
            )
    if json:
        if banner:
            rich.print_json(data=banners, indent=4)

        else:
            processed_items = {}

            for key, value in items.items():
                item_values = []

                for item in value:
                    item_values.append(item)
                processed_items[key] = item_values

            if title is not None:
                assert title in processed_items.keys(), (
                    f"Title filter '{title}' is not one of {list(processed_items.keys())}"
                )

                rich.print_json(data={title: processed_items.get(title)}, indent=4)
            else:
                rich.print_json(data=processed_items, indent=4)
    else:
        if banner:
            for key in banners.keys():
                target_banner = banners[key]
                table = Table(
                    title=f"{key} - Banner",
                    show_lines=True,
                )
                table.add_column("Pos")
                table.add_column("Subject type", style="white")  # justify="center")
                table.add_column("Title", style="cyan")

                table.add_column("Genre")
                table.add_column("Release date")
                table.add_column("IMDB Rating")

                for pos, item in enumerate(target_banner, start=1):
                    item.insert(0, str(pos))
                    table.add_row(*item)

                rich.print(table)

        else:
            if title is not None:
                target_title = items.get(title)
                assert target_title is not None, f"Title filter '{title}' is not one of {list(items.keys())}"
                items = {title: target_title}

            for key in items.keys():
                target_item = items[key]
                table = Table(
                    title=f"{key}",
                    show_lines=True,
                )
                table.add_column("Pos")
                table.add_column("Subject type", style="white")
                table.add_column("Title")

                table.add_column("Genre")
                table.add_column("IMDB Rating")

                table.add_column("Country name")
                table.add_column("Release date")

                for pos, item in enumerate(target_item, start=1):
                    item.insert(0, str(pos))
                    table.add_row(*item)

                rich.print(table)


@click.command(context_settings=command_context_settings)
@click.option(
    "-J",
    "--json",
    is_flag=True,
    help="Output details in json format : False",
)
def popular_search_command(json: bool):
    """Movies/tv-series many people are searching now"""
    prepare_start()

    search = PopularSearch(Session())
    items = get_event_loop().run_until_complete(search.get_content_model())

    if json:
        processed_items = [item.title for item in items]
        rich.print_json(data=dict(popular=processed_items), indent=4)

    else:
        table = Table(title="Popular Searches Now", show_lines=True)

        table.add_column("Pos")
        table.add_column("Title")

        for pos, item in enumerate(items, start=1):
            table.add_row(str(pos), item.title)

        rich.print(table)


@click.command(context_settings=command_context_settings)
@click.argument("title")
@click.option(
    "-y",
    "--year",
    type=click.INT,
    help="Year filter for the item to proceed with",
    default=0,
    show_default=True,
)
@click.option(
    "-s",
    "--subject-type",
    type=click.Choice(SubjectType.map().keys()),
    help="Item subject-type filter",
    default=SubjectType.ALL.name,
    show_default=True,
)
@click.option(
    "-Y",
    "--yes",
    is_flag=True,
    help="Do not prompt for item confirmation",
)
@click.option(
    "-J",
    "--json",
    is_flag=True,
    help="Output details in json format instead of tabulated",
)
@click.option("-F", "--full", is_flag=True, help="Show all details of the item")
@click.option(
    "-V",
    "--verbose",
    count=True,
    help="Show more detailed interactive texts",
    default=0,
)
@click.option(
    "-Q",
    "--quiet",
    is_flag=True,
    help="Disable showing interactive texts on the progress (logs)",
)
def item_details_command(json: bool, full: bool, verbose: int, quiet: bool, **item_kwargs):
    """Show details of a particular movie/tv-series"""
    prepare_start(quiet=quiet, verbose=verbose)

    item_kwargs["subject_type"] = getattr(SubjectType, item_kwargs.get("subject_type"))
    session = Session()

    target_item = get_event_loop().run_until_complete(
        perform_search_and_get_item(session=session, **item_kwargs)
    )

    subject_type_item_details_map = {
        SubjectType.MOVIES: MovieDetails,
        SubjectType.TV_SERIES: TVSeriesDetails,
    }

    ItemDetails: MovieDetails | TVSeriesDetails = subject_type_item_details_map.get(target_item.subjectType)

    assert ItemDetails, (
        f"The selected item type - {target_item.subjectType.name} - is not yet supported. "
        "Choose items of subject types "
        f"{' or '.join([key.name for key in list(subject_type_item_details_map.keys())])}"
    )

    item_details = ItemDetails(target_item, session=session)

    extractor = JsonDetailsExtractor(item_details.get_html_content_sync())

    details = target_item.model_dump() if full else {}

    details.update(extractor.metadata)

    season_items = []

    for season in extractor.seasons:
        season_string = (
            f"Season: {season['se']} "
            f"Episodes: {season['maxEp']} "
            f"Resolutions: {[res['resolution'] for res in season['resolutions']]}"
        )
        season_items.append(season_string)

    details["seasons"] = season_items

    if json:
        rich.print_json(data=details, indent=4)

    else:
        table = Table("Key", "Value", title=f"{details['title']} - details", show_lines=True)

        for key, value in details.items():
            table.add_row(key, "\n".join(value) if type(value) is list else str(value))

        rich.print(table)



"""Tests for timeout and retry CLI configuration."""

import subprocess

import httpx
import pytest

from moviebox_api.constants import (
    DEFAULT_CONNECT_RETRIES,
    DEFAULT_CONNECT_TIMEOUT,
    DEFAULT_POOL_TIMEOUT,
    DEFAULT_READ_TIMEOUT,
    DEFAULT_RETRY_BACKOFF,
    DEFAULT_WRITE_TIMEOUT,
)
from moviebox_api.requests import Session


def run_system_command(command: str) -> subprocess.CompletedProcess:
    return subprocess.run(
        "python -m moviebox_api " + command,
        shell=True,
        text=True,
        capture_output=True,
    )


class TestDefaultConstants:
    def test_default_connect_timeout(self):
        assert DEFAULT_CONNECT_TIMEOUT == 20.0

    def test_default_read_timeout(self):
        assert DEFAULT_READ_TIMEOUT == 60.0

    def test_default_write_timeout(self):
        assert DEFAULT_WRITE_TIMEOUT == 20.0

    def test_default_pool_timeout(self):
        assert DEFAULT_POOL_TIMEOUT == 20.0

    def test_default_connect_retries(self):
        assert DEFAULT_CONNECT_RETRIES == 5

    def test_default_retry_backoff(self):
        assert DEFAULT_RETRY_BACKOFF == 0.6


class TestTimeoutObjectCreation:
    def test_httpx_timeout_construction(self):
        timeout = httpx.Timeout(
            connect=DEFAULT_CONNECT_TIMEOUT,
            read=DEFAULT_READ_TIMEOUT,
            write=DEFAULT_WRITE_TIMEOUT,
            pool=DEFAULT_POOL_TIMEOUT,
        )
        assert timeout.connect == DEFAULT_CONNECT_TIMEOUT
        assert timeout.read == DEFAULT_READ_TIMEOUT
        assert timeout.write == DEFAULT_WRITE_TIMEOUT
        assert timeout.pool == DEFAULT_POOL_TIMEOUT

    def test_custom_timeout_construction(self):
        timeout = httpx.Timeout(connect=30.0, read=120.0, write=30.0, pool=30.0)
        assert timeout.connect == 30.0
        assert timeout.read == 120.0

    def test_transport_with_retries(self):
        transport = httpx.AsyncHTTPTransport(retries=DEFAULT_CONNECT_RETRIES)
        assert transport is not None

    def test_session_accepts_timeout_and_transport(self):
        timeout = httpx.Timeout(
            connect=DEFAULT_CONNECT_TIMEOUT,
            read=DEFAULT_READ_TIMEOUT,
            write=DEFAULT_WRITE_TIMEOUT,
            pool=DEFAULT_POOL_TIMEOUT,
        )
        transport = httpx.AsyncHTTPTransport(retries=DEFAULT_CONNECT_RETRIES)
        session = Session(timeout=timeout, transport=transport)
        assert session._timeout == timeout
        assert session._httpx_kwargs.get("transport") is transport

    def test_session_stores_httpx_kwargs(self):
        transport = httpx.AsyncHTTPTransport(retries=3)
        session = Session(transport=transport)
        assert session._httpx_kwargs.get("transport") is transport


@pytest.mark.parametrize(
    "command",
    [
        "download-movie --help",
        "download-series --help",
    ],
)
def test_timeout_options_in_help(command):
    result = run_system_command(command)
    assert result.returncode == 0
    output = result.stdout
    assert "--connect-timeout" in output
    assert "--read-timeout" in output
    assert "--write-timeout" in output
    assert "--pool-timeout" in output
    assert "--connect-retries" in output
    assert "--retry-backoff" in output
